#!/usr/bin/env python3

###################################################################################################
# ＜設定＞
#
# ホームディレクトリ =========================================================================
# このファイルを置くディレクトリ
homedir = '/home/pi/camera/'
#
# home.pyの実行Raspberry PiのIPアドレスとポート番号 ==========================================
# 別のRaspberry Piにファイル名を送信する場合、homeenableをTrueにして、送信先のIPアドレスとポート番号を設定します。
homeenable = False
homeipaddress = '192.168.1.10'
homeportaddress = 54321
#
# LINE通知 ==============================================================================
# カメラ検知したときにLINEグループへ通知する場合、lineenableをTrueにして、line.shをホームディレクトリに置きます。
lineenable = False
#
# その他 ================================================================================
SLEEPTIME = 1.1                # Sleep time in normal run. 1.1S.
THRESHOLD = 30                 # Difference detection threshold (~255)
WIDTH = 640                    # Width for analysis.
HEIGHT = 480                   # Height for analysis.
DETECTSIZE = WIDTH*HEIGHT/300  # Detection size criteria.
FRAMERATE = 30                 # Fixed frame rate.
RECINTERVAL = 60               # Interval of recording. 1min.
MAXRECTIME = 300               # Max recording number.
MAXTEMP = 65                   # Max temperature.

###################################################################################################
# インクルードファイル
import sys
import syslog
import cv2 as cv
import numpy as np
import subprocess
import picamera
import picamera.array
import datetime
import concurrent.futures
import glob
import RPi.GPIO as GPIO
from time import sleep
from fractions import Fraction
import socket
import os
import time

###################################################################################################
# ファイル名の送信
def sendfilename(str):
    global homeipaddress, homeportaddress
    filename = str.encode('utf-8')
    try:
        camerasocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        camerasocket.connect((homeipaddress, homeportaddress))
        camerasocket.send(filename)
    except Exception as e:
        print(e)
    finally:
        camerasocket.close()

###################################################################################################
# Piカメラ初期化
camera = picamera.PiCamera()
camera.resolution = (WIDTH, HEIGHT)
camera.framerate = FRAMERATE
camera.iso = 0
camera.saturation = 0
camera.rotation = 180
camera.awb_mode = 'greyworld'
stream = picamera.array.PiRGBArray(camera)

###################################################################################################
# OpenCV Mat (numpy array)
preimg = np.zeros([HEIGHT, WIDTH, 3], np.uint8)     # Previous image
kernel = np.ones((5,5),np.uint8)                    # Morphology

###################################################################################################
# Others初期化
slptime = SLEEPTIME    # Sleep time
imgfilename = ''       # Image file name
numimg = 1             # Serial number for image file name
recording = False      # Recording flag
nodetectcounter = 0    # Recording interval counter
timercounter = 0       # Timer
now = datetime.datetime.now()
datenow = now.strftime('%Y%m%d')
datepre = datenow
sn = 1                 # Serial number 1-9999 in one day
emergency = False      # Emergent Stop
night = 0              # Night flag 0:day 1:evening 2:night
nightcounter = 0       # Night Check Counter
searchfile = homedir+"/large/" + now.strftime('%Y%m%d') + "*.jpg"
files = glob.glob(searchfile)
max = 0
for filename in files:
    num = int(filename[32:36])
    if max < num:
        max = num
if max > sn:
    sn = max + 1
    if sn > 9999:
        sn = 9999
        emergency = True
GPIO.setmode(GPIO.BCM)
GPIO.setup(4,GPIO.IN)
gpiodetect = False

###################################################################################################
# LINE通知シェル実行
def sendmessage(img_name, imgarray):
    global lineenable, homedir, homeenable
    cv.imwrite(img_name, imgarray)
    if homeenable:
        sendfilename(os.path.basename(img_name))
    if lineenable:
        subprocess.run([homedir+'/line.sh', img_name ], encoding='utf-8', stdout=subprocess.DEVNULL)

###################################################################################################
# タイムラプスシェル実行
def maketimelapse(tl_name):
    global homedir
    subprocess.run([homedir+'/timelapse.sh', tl_name], encoding='utf-8', stdout=subprocess.DEVNULL)

###################################################################################################
# イメージ保存
def saveimage(img_name, imgarray):
    cv.imwrite(img_name, imgarray)

###################################################################################################
# メインプログラム
try:
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
    camera.start_preview()
    sleep(5)
    camera.capture(stream, 'bgr', use_video_port=False)
    preimg = cv.blur(stream.array, (5,5))
    syslog.syslog(syslog.LOG_INFO, 'start')
    while emergency == False:
        if slptime > 0:
            sleep(slptime)
        timercounter+=slptime+1
        stream.truncate(0)
        gpiodetect = GPIO.input(4)
        camera.capture(stream, 'bgr', use_video_port=False)
        imgblur = cv.blur(stream.array, (5,5))
        diff = cv.absdiff(preimg,imgblur)
        ret, th = cv.threshold(diff, THRESHOLD, 255, cv.THRESH_BINARY)
        mor = cv.morphologyEx(th, cv.MORPH_CLOSE, kernel)
        grayimg = cv.cvtColor(mor, cv.COLOR_BGR2GRAY)
        contours, hierarchy = cv.findContours(grayimg, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
        contoursizemax = 0
        contouri = 0
        i = 0
        for contour in contours:
            contoursize = cv.contourArea(contour)
            if contoursize > contoursizemax:
                contoursizemax = contoursize
                contouri = i
            i+=1
        if contoursizemax > DETECTSIZE and gpiodetect == True:   # Detect
            nodetectcounter = 0
            dtstamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            cv.putText(stream.array, dtstamp, (402,22), cv.FONT_HERSHEY_SIMPLEX, 0.5, (0,0,0), 1, cv.LINE_AA)
            cv.putText(stream.array, dtstamp, (400,20), cv.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1, cv.LINE_AA)
            if not recording:
                now = datetime.datetime.now()
                recording = True
                numimg = 1
                contimg = cv.drawContours(stream.array, contours, contouri, (0,255,0), 1)
            imgfilename = homedir + '/large/' + now.strftime('%Y%m%d_') + '%04d.jpg' % sn
            sn+=1
            if sn > 9999:
                recording = False
                emergency = True
                syslog.syslog(syslog.LOG_ERR, 'file nu,mber over')
            cnt=contours[contouri]
            rightmost = tuple(cnt[cnt[:,:,0].argmax()][0])
            if numimg == 1 and rightmost[0] < 600:
                executor.submit(sendmessage, imgfilename, contimg)
            else:
                executor.submit(saveimage, imgfilename, stream.array)
            numimg+=1
            if numimg > MAXRECTIME:                              # Emergent stop
                recording = False
                emergency = True
                syslog.syslog(syslog.LOG_ERR, 'recording time over')
            elif numimg > 30:                                    # Slowest capture
                slptime = 10
            elif numimg > 10:                                    # Slower capture
                slptime = 2
            else:
                slptime = 0
        elif recording:
            nodetectcounter+=slptime+1
            if numimg > 30:                                      # Slowest capture
                slptime = 10
            elif numimg > 10:                                    # Slower capture
                slptime = 2
            else:
                slptime = 0
            if nodetectcounter > RECINTERVAL:                    # Finish recording
                nodetectcounter = 0
                recording = False
                slptime = SLEEPTIME
        preimg = imgblur.copy()

        if timercounter > 60 and not recording:
            sum = cv.sumElems(stream.array)[0]
            if sum < 500000 and night == 0:
                night = 1
                syslog.syslog(syslog.LOG_INFO, 'Night detected.')
                camera.iso = 800
                camera.framerate = 3
                camera.awb_mode = 'off'
                camera.awb_gains = (Fraction(8, 1), Fraction(8, 1))
                camera.shutter_speed = 300000
                camera.saturation = 100
            elif sum > 60000000 and  night == 1:
                camera.awb_mode = 'auto'
                night = 0
                syslog.syslog(syslog.LOG_INFO, 'Morning detected.')
                camera.iso = 0
                camera.framerate = FRAMERATE
                camera.awb_mode = 'greyworld'
                camera.shutter_speed = 0
                camera.saturation = 0
            cp = subprocess.run(['vcgencmd', 'measure_temp'], encoding='utf-8', stdout=subprocess.PIPE)
            tempnow = float(cp.stdout[5:9])
            if tempnow > MAXTEMP:
                emergency = True
                syslog.syslog(syslog.LOG_ERR, 'over temperature')
            now2 = datetime.datetime.now()
            datepre = datenow
            datenow = now2.strftime('%Y%m%d')
            if datepre != datenow:
                executor.submit(maketimelapse, datepre)
                sn=1
            timercounter = 0
    camera.stop_preview()
finally:
    GPIO.cleanup(4)
    camera.close()
    executor.shutdown()
    syslog.syslog(syslog.LOG_INFO, 'end')
